import java.util.LinkedList;
import java.util.ListIterator;

/**

 The FlowGraph class represents an instance of the MAX-FLOW problem. Other than
 the constructor, the two important public methods are printCapacityMatrix(),
 which pretty prints the capacity matrix of the flow graph, and calculateFlow(),
 which returns the maximum flow of the graph, and also modifies the graph
 according to the standard MAX-FLOW algorithm.

 @author SJSU Students

 */
public class FlowGraph 
{

	private Node[] uNodes, vNodes; // These arrays will have the same length.
	private Node source, sink;
	private EdgeList edges;

	/**
	 This is a constructor for the class. It simply takes in the list of
	 strings passed to Matching.main(), which represent a bipartite graph, and
	 converts the strings into the internal representation of a graph.
	 @param edgeLists strings representing an instance of BIPARTITE-MATCHING
	 @return the flow graph resulting from a reduction to MAX-FLOW
	 */
	public FlowGraph(String[] edgeLists)
	{
		uNodes = new Node[edgeLists.length];
		vNodes = new Node[edgeLists.length];

		edges = new EdgeList();

		// Initialize all of the nodes. The numbering is defined in the
		// assignment.
		for(int index = 0; index < edgeLists.length; index++)
		{
			uNodes[index] = new Node(index);
			vNodes[index] = new Node(index + edgeLists.length);
		}
		source = new Node(edgeLists.length * 2);
		sink = new Node((edgeLists.length * 2) + 1);

		// Add the edges defined by the input strings.
		for(int uIndex = 0; uIndex < edgeLists.length; uIndex++)
			addEdges(uIndex, edgeLists[uIndex]);

		// Add an edge from the source to all the U nodes, and from all the V
		// nodes to the sink.
		for (int index = 0; index < edgeLists.length; index++)
		{
			linkNodes(source, uNodes[index]);
			linkNodes(vNodes[index], sink);
		}
	}

	/**
	 This method adds edges from the U node specified by uNumber to the V nodes
	 listed in neighborString.
	 @param uNumber the number of the U node
	 @param neighborString a string containing the V nodes to which the U node
	 is connected
	 */
	private void addEdges(int uNumber, String neighborString)
	{
		Node uNode = uNodes[uNumber];
		int[] neighbors = parseIntString(neighborString);

		for (int i = 0; i < neighbors.length; i++)
			linkNodes(uNode, vNodes[neighbors[i]]);
	}

	/**
	 This method creates a new edge linking the two input nodes, and adds the
	 edge to the edge list in both the current graph as well as the edge list
	 in each of the two nodes.
	 @param from the node which will be on the "from" side of the new edge
	 @param to the node which will be on the "to" side of the new edge
	 */
	private void linkNodes(Node from, Node to)
	{
		Edge e = new Edge(from, to);
		from.addEdge(e);
		to.addEdge(e);
		edges.add(e);
	}

	/**
	 This method parses s into an array of integers. s is assumed to be a string
	 containing integers separated by single spaces.
	 @param s the string of integers to be parsed
	 @return the array of integers resulting from parsing the string
	 */
	private int[] parseIntString (String s)
	{
		if (s.equals("")) return new int[0];  // Just in case.

		String[] intStringArray = s.split(" ");
		int[] returnArray = new int[intStringArray.length];

		for (int i = 0; i < returnArray.length; i++)
			returnArray[i] = Integer.decode(intStringArray[i]).intValue();

		return returnArray;
	}

	/**
	 This method pretty prints to the screen the capacity matrix of the current
	 flow graph. All edge capacities are assumed to be 1.
	 */
	public void printCapacityMatrix()
	{
		// First, print the rows corresponding to the U edges.
		for(int uIndex = 0; uIndex < uNodes.length; uIndex++)
			System.out.println(constructEdgeString(uNodes[uIndex]));

		// Next, print the rows corresponding to the V edges.
		for (int vIndex = 0; vIndex < vNodes.length; vIndex++)
			System.out.println(constructEdgeString(vNodes[vIndex]));

		// Finally, print the rows for the source and the sink, in that order.
		System.out.println(constructEdgeString(source));
		System.out.println(constructEdgeString(sink));
	}

	/**
	 This method constructs a string corresponding to the outgoing edges of
	 inNode. A 1 in the kth column indicates that inNode has an edge to node k,
	 and a 0 indicates that it does not. (The numbering of the nodes is defined
	 in the assignment.)
	 @param inNode the node whose edge string is to be constructed
	 @return the string of 1's and 0's representing the outgoing edges of inNode
	 */
	private String constructEdgeString(Node inNode)
	{
		String s = "";  // This will be the string that we return.

		// The first columns correspond to the U nodes.
		for (int uIndex = 0; uIndex < uNodes.length; uIndex++)
			s += edges.contains(new Edge(inNode, uNodes[uIndex])) ? " 1" : " 0";

		// The next columns correspond to the V nodes.
		for (int vIndex = 0; vIndex < vNodes.length; vIndex++)
			s += edges.contains(new Edge(inNode, vNodes[vIndex])) ? " 1" : " 0";

		// The final two columns correspond to the source and the sink,
		// respectively.
		s += edges.contains(new Edge(inNode, source)) ? " 1" : " 0";
		s += edges.contains(new Edge(inNode, sink)) ? " 1" : " 0";

		return s;
	}

	/**
	 This method calculates the maximum flow in the network and returns it.
	 Note that this will cause the  graph to be modified as per the standard
	 MAX-FLOW algorithm.
	 @return the maximum flow of this flow graph
	 */
	public int calculateFlow()
	{
		int flow = 0;  // Initialize the flow to 0.

		// Find the first path from source to sink.
		EdgeList currentPath = sourceToSinkPath();

		// Keep incrementing the flow while we still have a path from
		// the source to the sink in this graph.
		while (currentPath != null)
		{
			// Since there is a path, we should increment the flow by the
			// smallest capacity on the path; however, since all edges have
			// capacity 1, we just increment the flow by 1.
			flow++;

			// For each edge in the path, remove it from this graph, and add
			// its reverse.
			ListIterator currentPathIterator = currentPath.listIterator();
			while (currentPathIterator.hasNext())
			{
				Edge e = (Edge) currentPathIterator.next();
				removeEdge(e);
				addReverseEdge(e);
			}

			// Find the next path.
			currentPath = sourceToSinkPath();
		}

		return flow;
	}

	/**
	 This method adds the reverse of the input edge to this graph and to the
	 edge lists of the two nodes on the edge.
	 @param e the edge whose reverse is to be added to this graph
	 */
	private void addReverseEdge(Edge e)
	{
		Edge reverseEdge = e.reverse();
		edges.add(reverseEdge);
		e.getFromNode().addEdge(reverseEdge);
		e.getToNode().addEdge(reverseEdge);
	}

	/**
	 This method removes an edge from this graph and from the edge lists of the
	 two nodes on the edge.
	 @param e the edge to be removed
	 */
	private void removeEdge(Edge e)
	{
		edges.remove(e);
		e.getFromNode().removeEdge(e);
		e.getToNode().removeEdge(e);
	}

	/**
	 This method finds a path from the source to the sink (using breadth-first
	 search), and returns it as an ordered list of nodes representing the path.
	 If no path exists, null is returned.
	 @return the ordered list of edges on a path from the source to the sink
	 */
	private EdgeList sourceToSinkPath()
	{
		LinkedList nodes = new LinkedList();

		// This array will help us reconstruct the path, assuming one exists.
		// If node k is at index i, this means that there is an edge from
		// node k to node i; specifically, the edge from k to i is the edge we
		// used to reach i.
		Node[] pathArray = new Node[(2 * uNodes.length) + 2];

		// Unmark all the nodes from any previous searches.
		unmarkAllNodes();

		// Initially, only the source is in the list.
		source.mark();
		nodes.add(source);

		while (!nodes.isEmpty())
		{
			// Pull the first node from the front of the queue.
			Node currentNode = (Node) nodes.getFirst();
			nodes.removeFirst();

			// Loop through the outgoing edges of the current node, and add
			// all unmarked neighbors to the queue, while also marking them and
			// using pathArray to note how we reached them.
			ListIterator currentEdges = currentNode.outgoingEdgesIterator();
			while (currentEdges.hasNext())
			{
				Edge e = (Edge) currentEdges.next();
				if (!e.getToNode().isMarked())
				{
					e.getToNode().mark();
					nodes.addLast(e.getToNode());  // Adding to the end.
					pathArray[e.getToNode().getNumber()] = currentNode;
				}
			}
		}

		// If we didn't reach the sink, then there is no path, so return null.
		if (!sink.isMarked()) return null;

		// Otherwise, reconstruct the path in reverse using pathArray.
		EdgeList path = new EdgeList();

		// First, we add the edge that goes to the sink.
		path.add(new Edge(pathArray[sink.getNumber()], sink));

		// Keep looping until the first edge in our path starts at the source.
		Node front = ((Edge) path.getFirst()).getFromNode();
		while (!front.equals(source))
		{
			path.addFirst(new Edge(pathArray[front.getNumber()], front));
			front = pathArray[front.getNumber()];
		}

		return path;
	}

	/**
	 This method simply calls unmark() on each node in this graph.
	 */
	private void unmarkAllNodes()
	{
		// This takes advantage of the fact that uNodes.length == vNodes.length.
		for (int index = 0; index < uNodes.length; index++)
		{
			uNodes[index].unmark();
			vNodes[index].unmark();
		}
		source.unmark();
		sink.unmark();
	}
}

